#include "pointer.h"   // Pointer class prototype
#include <stdio.h>     // NULL
#include <iostream.h>  // cerr

extern void error(char *);      // error 

template<class T>
Pointer<T> :: Pointer(void)
{
  // initilize
  p_info = NULL;
  ptr = NULL;
  left = right = NULL;
  index = 0;
}

template<class T>
Pointer<T> :: Pointer(Pointer& p)
{  
  // copy general information
  p_info = p.p_info;  
  index = p.index;
  ptr = p.ptr;
  left = right = NULL;
  
  // exclude auto variable
  if (p.p_info != NULL) {   
    
    // one more pointer to the allocated memory
    
    (p_info->counter)++;  // increase the number of pointers
    
    // link into the link list
    left = &p; right = p.right; 
    if (p.right != NULL) (*(p.right)).left = this;
    p.right = this;
    
  };
}

template<class T>
Pointer<T> :: Pointer(T* tptr) 
{ 
  // initilize and copy
  index = 0; p_info = NULL; 
  left = right = NULL; 
  ptr = tptr;

 }

template<class T>
Pointer<T> :: ~Pointer(void)
{  
  if (p_info != NULL) this->deleteptr();
}

template<class T>
Pointer<T>& Pointer<T> :: deleteptr(void)
{ 
  if (p_info != NULL) {
    if (--(p_info->counter) == 0) {  // no more pointer to this memory
      // delete the allocated memory
      ptr -= index;  // move to pointer to the first element 
      delete [] ptr; 
      // delete the space used for general information
      delete p_info;
    }
    else {  // there are other pointer(s) to this memory
      if (left != NULL) left->right = right;
      if (right != NULL) right->left = left;
    };
    // re-initilize
    p_info = NULL; left = right = NULL; index = 0;
  };

  return *this;
}      

template<class T>
T& Pointer<T> :: operator *(void) 
{   
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) 
    cerr << "Warning : Bad reference, outside of bound \n";
  
  return *ptr; 
}

template<class T>
T* Pointer<T> :: operator->(void) 
{ 
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {
    cerr << "Warning : Bad reference, outside of bound \n";
    return NULL;
  }
  else  
    return ptr; 
}

template<class T>
T& Pointer<T> :: operator[](const int i) 
{   
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0)))
    cerr << "Warning : outside of bound \n";

  return ptr[i-index];
}

template<class T>
Pointer<T> Pointer<T> :: operator+(const int n)
{  
  int i = index + n;
  
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {
    cerr << "Warning : outside of bound \n";
    return NULL;
  };
  
  Pointer temp(*this);  // copy to a temporary object
  temp.index = i; (temp.ptr) += n; 
  
  return temp;
}

template<class T>
Pointer<T> Pointer<T> :: operator-(const int n)
{  
  int i = index - n;

  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {
    cerr << "Warning : outside of bound \n";
    return NULL;
  };
  
  Pointer temp(*this); // copy to a temporiary object
  temp.index = i; (temp.ptr) -= n;
  
  return temp;
}

// prefix operation ++ptr
template<class T>
Pointer<T>& Pointer<T> :: operator++(void)
{
  index++;
  ptr++;

  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return *this;
}

// postfix ptr++
template<class T>
Pointer<T> Pointer<T> :: operator++(int)
{  
  Pointer save(*this);
  
  index++;
  ptr++;
  
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {  
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return save;
}

// --ptr
template<class T>
Pointer<T>& Pointer<T> :: operator--(void)
{  
  index--;
  ptr--;

  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {  
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return *this;
}

// ptr--
template<class T>
Pointer<T> Pointer<T> :: operator--(int)
{  
  Pointer save(*this);
  
  index--;
  ptr--;
  
  // bound checking
  if ((p_info == NULL && index != 0) ||
      (p_info != NULL && (index >(p_info->size -1) || index < 0))) {  
    cerr << "Warning : outside of bound \n";
    return NULL;
  };

  return save;
}

template<class T>
Pointer<T>& Pointer<T> :: operator=(Pointer& p)
{  
  // release the memory previously pointed to
  if (p_info != NULL) this->deleteptr();
  
  // copy general information
  p_info = p.p_info;
  ptr = p.ptr;
  index = p.index;
  
  // exclude auto variable
  if (p.p_info != NULL) {   
    (p_info->counter)++;
    left = &p; right = p.right; 
    if (p.right != NULL) (*(p.right)).left = this;
    p.right = this;
  };
  
  return *this;
}

template<class T>
Pointer<T>& Pointer<T> :: operator=(T* tptr )
{
  // release the memory previously pointed to
  if (p_info != NULL) this->deleteptr();

  ptr = tptr; index = 0; p_info = NULL;
  left = right = NULL;
  
  return *this;
}

template<class T>
Pointer<T>& Pointer<T> :: newptr(const int n = 1)
{
#define MAXSIZE 50000  // maximum number of array elements 
  
  // release the memory previously pointed to
  if (p_info != NULL) this->deleteptr();  
  
  if (n > MAXSIZE || n < 1) error("error:incorrect size\n");

  p_info = new P_info;
  p_info->size = n;
  p_info->counter = 1;
  left = right = NULL;
  
  // allocate the object(s) memory
  
  ptr = new T[n];
  index = 0;
  
  return *this;
}


